/**
 * Copyright (c) 2010-2012 EBM WebSourcing, 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.orchestration;

import java.io.IOException;

import javax.activation.DataHandler;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessagingException;
import javax.xml.namespace.QName;
import javax.xml.transform.Source;

public interface Orchestrator {

	public static final int IN_ONLY = 0;
	public static final int IN_OUT = 1;
	public static final int ROBUST_IN_ONLY = 2;
	
	public static final int ERROR = 1;
	public static final int FAULT = 2;

	public int getErrorStatus();
	public void resetError();
	public Exception getError();
	public Fault getFault();

	public abstract Source getContent() throws MessagingException;

	public abstract Source getContent(boolean consume)
			throws MessagingException;

	public abstract String getContentAsString() throws Exception;

	public abstract DataHandler getAttachment(String name);

	public abstract Orchestrator putAttachment(String name, DataHandler attachment);

	public abstract DataHandler removeAttachment(String name);

	public abstract boolean hasAttachments();

	public abstract Orchestrator setOperation(QName operation);

	public abstract Orchestrator setOperation(String ns, String localpart);

	public abstract Orchestrator setOperation(String localpart);

	/**
	 * Specify no operation.
	 * @return Current orchestrator.
	 */
	public abstract Orchestrator clearOperation();

	/**
	 * Method dedicated to asynchronous handling of in-out calls, for synchronization:
	 * returns null, until the asynchronous in-out call has ended; then, returns the
	 * orchestrator related to the in-out response.
	 * Typical use:
	 *  Orchestrator result;
	 *  while((result = getCurrent()) == null); //wait
	 *  // Now use result in next orchestration
	 * @return The current orchestrator, to be used from now.
	 */
	public abstract Orchestrator getCurrent();

	public abstract Orchestrator consume(String epr) throws Exception;

	public abstract Orchestrator inOnly(String epr) throws Exception;
	public abstract Orchestrator robustInOnly(String epr) throws Exception;
	public abstract Orchestrator inOut(String epr) throws Exception;

	public abstract Orchestrator consumeAsync(String epr, int mep)
			throws Exception;

	/**
	 * Unzip a specified attachment of the current message.
	 * @return The updated message (with attachment unzipped).
	 * @throws Exception
	 */
	public abstract Orchestrator unzipAttachment(String attname)
			throws Exception;

	/**
	 * Unzip any zipped attachment of the current message.
	 * @return The updated message (with attachments unzipped).
	 * @throws Exception
	 */
	public abstract Orchestrator unzipAttachments() throws Exception;

	/**
	 * Switch attachment to payload (the previous payload gets lost, and the corresponding attachment is removed).
	 * When no attachment is eligible, nothing is done.
	 * @param attname The attachment name (null means 1st attachment).
	 * @return The possibly updated message.
	 * @throws Exception
	 */
	public abstract Orchestrator attachmentToPayload(String attname)
			throws Exception;

	/**
	 * Transform the message payload using XSLT
	 * @param xslt The XSL stylesheet (relative to current SU path).
	 * @return The transformed message.
	 */
	public abstract Orchestrator transform(String xslt) throws Exception;

	/**
	 * Transform the message payload using XSLT, and put result as attachment.
	 * @param xslt The XSL stylesheet (relative to current SU path)
	 * @param attname The attachment name
	 * @return Current orchestrator, with attachment transformed
	 */
	public abstract Orchestrator transform(String xslt, String attname)
			throws Exception;

	/**
	 * Write payload to a file.
	 * @param fileName File name or path
	 * @return Current orchestrator
	 * @throws IOException
	 */
	public abstract Orchestrator writePayload(String fileName)
			throws IOException;

	/**
	 * Write specified attachment to a file, the file name being set to the attachment name.
	 * @param path The output directory
	 * @param attname The attachment name
	 * @return Current orchestrator
	 * @throws IOException
	 */
	public abstract Orchestrator writeAttachment(String path, String attname)
			throws IOException;

	/**
	 * Write attachments into a directory, each file name being set to the corresponding attachment name.
	 * @param path The output directory
	 * @return Current orchestrator
	 * @throws IOException
	 */
	public abstract Orchestrator writeAttachments(String path)
			throws IOException;

	/**
	 * Write payload + attachments into a directory
	 * @param path The output directory
	 * @param payloadName The file name for the payload
	 * @return Current orchestrator
	 * @throws IOException
	 */
	public abstract Orchestrator write(String path, String payloadName)
			throws IOException;

	public abstract Orchestrator push();

	public abstract Orchestrator pop();

}
