/**
 * Copyright (c) 2010-2012 EBM WebSourcing, 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.orchestration.core;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringReader;
import java.util.Set;

import javax.activation.DataHandler;
import javax.jbi.component.ComponentContext;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessagingException;
import javax.xml.namespace.QName;
import javax.xml.transform.Source;
import javax.xml.transform.stream.StreamSource;

import org.ow2.easywsdl.wsdl.api.abstractItf.AbsItfOperation.MEPPatternConstants;
import org.ow2.petals.component.framework.api.Message.MEPConstants;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.listener.AbstractJBIListener;
import org.ow2.petals.orchestration.Orchestrator;

public class OrchestratorImpl implements Orchestrator {
	
	private final InvocationContext iCtx_;
	private final AbstractJBIListener jbiListener_;
	private final ComponentContext ctx_;

    private final String installRoot_;
	
	private OrchestratorImpl lastExchanged_ = this;
	private boolean isDone_ = true;
	
	private int errorStatus_ = 0;
	private Exception error_ = null;
	private Fault fault_ = null;
	
	private OrchestratorImpl parent_ = null;
	
    public OrchestratorImpl(AbstractJBIListener listener, ComponentContext ctx, InvocationContext iCtx,
            String installRoot) {
		jbiListener_ = listener;
		ctx_ = ctx;
        iCtx_ = iCtx;
        this.installRoot_ = installRoot;
	}

    public OrchestratorImpl(String content, AbstractJBIListener listener, ComponentContext ctx, String installRoot) {
        this(content, IN_OUT, listener, ctx, installRoot);
	}

    public OrchestratorImpl(String content, int mep, AbstractJBIListener listener, ComponentContext ctx,
            String installRoot) {
        this(listener, ctx, new InvocationContext(), installRoot);
		iCtx_.setContent(new StreamSource(new StringReader(content)));

		if(mep == IN_ONLY) iCtx_.setMep(MEPConstants.IN_ONLY_PATTERN);
		else if(mep == ROBUST_IN_ONLY) iCtx_.setMep(MEPConstants.ROBUST_IN_ONLY_PATTERN);
		else iCtx_.setMep(MEPConstants.IN_OUT_PATTERN);
	}

    public OrchestratorImpl(Exchange e, AbstractJBIListener listener, ComponentContext ctx, String installRoot)
            throws Exception {
        this(listener, ctx, new InvocationContext(e), installRoot);
	}
	
	public AbstractJBIListener getJbiListener() { return jbiListener_; }

	public InvocationContext getInvocationContext() throws MessagingException {
		return iCtx_;
	}

	@Override
	public Source getContent() throws MessagingException {
		return getContent(true);
	}

	@Override
	public Source getContent(boolean consume) throws MessagingException {
		return (lastExchanged_ == this ? iCtx_.getContent(consume) : lastExchanged_.getContent(consume));
	}

	@Override
	public String getContentAsString() throws Exception {
		return OrchestrationUtils.sourceToString(getContent(false));
	}

	@Override
	public DataHandler getAttachment(String name) { return iCtx_.getAttachment(name); }

	@Override
	public Orchestrator putAttachment(String name, DataHandler attachment) {
		iCtx_.putAttachment(name, attachment);
		return this;
	}

	@Override
	public DataHandler removeAttachment(String name) { return iCtx_.removeAttachment(name); }

	@Override
	public boolean hasAttachments() { return iCtx_.hasAttachments(); }

	@Override
	public Orchestrator setOperation(QName operation) {
		iCtx_.setOperation(operation);
		return this;
	}

	@Override
	public Orchestrator setOperation(String ns, String localpart) {
		if(ns == null) ns = "";
		iCtx_.setOperation(localpart != null ? new QName(ns, localpart) : null);
		return this;
	}

	@Override
	public Orchestrator setOperation(String localpart) {
		String ns = (iCtx_.getOperation() == null ? "" : iCtx_.getOperation().getNamespaceURI());
		iCtx_.setOperation(new QName(ns, localpart));
		return this;
	}

	@Override
	public Orchestrator clearOperation() {
		iCtx_.setOperation(null);
		return this;
	}
	
	@Override
	public int getErrorStatus() { return errorStatus_; }
	@Override
	public Exception getError() { return error_; }
	@Override
	public Fault getFault() { return fault_; }
	@Override
	public void resetError() {
		errorStatus_ = 0;
		fault_ = null;
		error_ = null;
	}
	
	protected Orchestrator consume(String epr, MEPPatternConstants mep, QName operation) throws Exception {
		return consume(epr, mep, operation, true);
	}
	
	/**
	 * Consumes a specified Endpoint
	 * @param epr The endpoint to consume
	 * @param mep The MEP (generally InOnly or InOut)
	 * @param operation The requested operation (null means forward current operation).
	 * @param synchronous Call in synchronous (true) or asynchronous mode
	 * @return The result when invoked InOut, and the incoming message when invoked InOnly.
	 */
	protected Orchestrator consume(String epr, MEPPatternConstants mep, QName operation, boolean synchronous)
	throws Exception {
		
		//TBD
		//if(lastExchanged_ != null && lastExchanged_.getFault() != null) return lastExchanged_;
		
		if(jbiListener_ == null) throw new Exception("JBI Listener null (not yet injected ?)");
		
		if(mep != null) iCtx_.setMep(mep);
		if(operation != null) iCtx_.setOperation(operation);
		
		Exchange out = iCtx_.createConsumeExchange(epr, jbiListener_);

		resetError();

		if(synchronous) {
			sendExchange(out, false, null);
			if(mep == MEPConstants.IN_ONLY_PATTERN || mep == MEPConstants.ROBUST_IN_ONLY_PATTERN) {
				return this;
			} else {
                lastExchanged_ = new OrchestratorImpl(out, jbiListener_, ctx_, installRoot_);
				return lastExchanged_;
			}
		} else { // asynchronous call
			
			if(mep == MEPConstants.IN_ONLY_PATTERN || mep == MEPConstants.ROBUST_IN_ONLY_PATTERN) {
				//OrchestrationAsyncContext ctx = new OrchestrationAsyncContext(iCtx_.getOriginalExchange());
				sendExchange(out, true, null);
				jbiListener_.send(out);
				return this;
			} else {
				OrchestratorImpl asyncBranch = fork();
				OrchestrationAsyncContext ctx = new OrchestrationAsyncContext(asyncBranch.getInvocationContext().getOriginalExchange());
				asyncBranch.setCurrent(false, null);
				ctx.setOriginalOrchestrator(asyncBranch);
				sendExchange(out, true, ctx);
				return asyncBranch;
			}
		}
	}

	private void sendExchange(Exchange out, boolean async, OrchestrationAsyncContext ctx) throws Exception {
		try {
			boolean timedOut = false;
			if(async) {
				if(ctx != null) jbiListener_.sendAsync(out, ctx);
				else jbiListener_.send(out);
			} else {
				timedOut = ! jbiListener_.sendSync(out);
			}
			if(timedOut) {
				errorStatus_ = Orchestrator.ERROR;
				error_ = new IOException("Petals: timeout when sending message");
				throw error_;
			} else if((fault_ = out.getFault()) != null) {
				errorStatus_ = Orchestrator.FAULT;
			}
		} catch(Exception e) {
			//System.out.println("Exception, status=" + out.getStatus() + ", error=" + out.getError() + ", fault=" + out.getFault());
			if(out.isErrorStatus()) {
				errorStatus_ = Orchestrator.ERROR;
				error_ = out.getError();
				if(error_ == null) {
					out.setError(e); // was missing...
					error_ = e;
				}
				throw error_;
			} else if((fault_ = out.getFault()) != null) { // Should never happen ?
				errorStatus_ = Orchestrator.FAULT;
				throw new Exception("Orchestration fault:\n"
						+ OrchestrationUtils.sourceToString(fault_.getContent()));
			} else {
				throw e;
			}
		}
	}

	@Override
	public synchronized Orchestrator getCurrent() {
		return (isDone_ ? lastExchanged_ : null);
	}

	public synchronized void setCurrent(boolean done, OrchestratorImpl o) {
		//System.out.println("setCurrent: " + this + " : " + done + ", " + o);
		if(o != null) lastExchanged_ = o;
		isDone_ = done;
	}
	
	@Override
	public Orchestrator consume(String epr) throws Exception {
		return consume(epr, null, null);
	}

	@Override
	public Orchestrator inOnly(String epr) throws Exception {
		return consume(epr, MEPConstants.IN_ONLY_PATTERN, null);
	}

	@Override
	public Orchestrator robustInOnly(String epr) throws Exception {
		return consume(epr, MEPConstants.ROBUST_IN_ONLY_PATTERN, null);
	}

	@Override
	public Orchestrator inOut(String epr) throws Exception {
		return consume(epr, MEPConstants.IN_OUT_PATTERN, null);
	}
	
	@Override
	public Orchestrator consumeAsync(String epr, int mep) throws Exception {
		switch(mep) {
			case IN_ONLY:
				return consume(epr, MEPConstants.IN_ONLY_PATTERN, null, false);
			case ROBUST_IN_ONLY:
				return consume(epr, MEPConstants.ROBUST_IN_ONLY_PATTERN, null, false);
			case IN_OUT:
			default:
				return consume(epr, MEPConstants.IN_OUT_PATTERN, null, false);
		}
	}
	
	@Override
	public Orchestrator unzipAttachment(String attname) throws Exception {
		iCtx_.unzipAttachment(attname);
		return this;
	}

	@Override
	public Orchestrator unzipAttachments() throws Exception {
		iCtx_.unzipAttachments();
		return this;
	}
	
	@Override
	public Orchestrator attachmentToPayload(String attname) throws Exception {
		iCtx_.attachmentToPayload(attname);
		return this;
	}
	
	@Override
	public Orchestrator transform(String xslt) throws Exception {
		iCtx_.transform(
				installRoot_ + File.separator + xslt,
				new SuURIResolver(installRoot_),
				null);
		return this;
	}
	
	@Override
	public Orchestrator transform(String xslt, String attname) throws Exception {
		iCtx_.transform(
				installRoot_ + File.separator + xslt,
				new SuURIResolver(installRoot_),
				attname);
		return this;
	}
	
	@Override
	public Orchestrator writePayload(String fileName) throws IOException {
		File f = new File(fileName);
		BufferedWriter out = new BufferedWriter(new FileWriter(f));
		try {
			String s = getContentAsString();
			out.write(s, 0, s.length());
		} catch (IOException e) {
			throw(e);
		} catch (Exception e) {
			throw new IOException(e);
		} finally {
			out.close();
		}
		return this;
	}
	
	@Override
	public Orchestrator writeAttachment(String path, String attname) throws IOException {
		DataHandler att = iCtx_.getAttachment(attname);
		InputStream in = att.getInputStream();
		byte buf[] = new byte[1024];
		FileOutputStream out = new FileOutputStream(path
				+ (path.endsWith(File.separator) ? "" : File.separatorChar)
				+ att.getName());
		while(in.read(buf) > 0) out.write(buf);
		out.close();
		return this;
	}
	
	@Override
	public Orchestrator writeAttachments(String path) throws IOException {
		Set<String> names = iCtx_.getAttachmentNames();
		for(String name : names) {
			writeAttachment(path, name);
		}
		return this;
	}
	
	@Override
	public Orchestrator write(String path, String payloadName) throws IOException {
		if(payloadName != null && payloadName.length() > 0)
			writePayload(path + (path.endsWith(File.separator) ? "" : File.separatorChar) + payloadName);
		writeAttachments(path);
		return this;
	}
	
	private OrchestratorImpl fork() {
        OrchestratorImpl next = new OrchestratorImpl(jbiListener_, ctx_, new InvocationContext(iCtx_, ctx_),
                installRoot_);
		next.setParent(parent_);
		return next;
	}

	@Override
	public Orchestrator push() {
        OrchestratorImpl next = new OrchestratorImpl(jbiListener_, ctx_, new InvocationContext(iCtx_, ctx_),
                installRoot_);
		next.setParent(this);
		return next;
	}

	@Override
	public Orchestrator pop() {
		return(parent_ == null ? this : parent_);
	}

	private void setParent(OrchestratorImpl o) { parent_ = o;}

}
